# Copyright 2023 Wieger Wesselink.
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE or http://www.boost.org/LICENSE_1_0.txt)

import jax.numpy as jnp

from nerva_jax.utilities import parse_function_call

Matrix = jnp.ndarray


def Relu(X: Matrix):
    return jnp.maximum(0, X)


def Relu_gradient(X: Matrix):
    return jnp.where(X > 0, 1, 0)


def Leaky_relu(alpha):
    return lambda X: jnp.maximum(alpha * X, X)


def Leaky_relu_gradient(alpha):
    return lambda X: jnp.where(X > 0, 1, alpha)


def All_relu(alpha):
    return lambda X: jnp.where(X < 0, alpha * X, X)


def All_relu_gradient(alpha):
    return lambda X: jnp.where(X < 0, alpha, 1)


def Hyperbolic_tangent(X: Matrix):
    return jnp.tanh(X)


def Hyperbolic_tangent_gradient(X: Matrix):
    return 1 - jnp.tanh(X) ** 2


def Sigmoid(X: Matrix):
    return 1 / (1 + jnp.exp(-X))


def Sigmoid_gradient(X: Matrix):
    return Sigmoid(X) * (1 - Sigmoid(X))


def Srelu(al, tl, ar, tr):
    return lambda X: jnp.where(X <= tl, tl + al * (X - tl),
                     jnp.where(X < tr, X, tr + ar * (X - tr)))


def Srelu_gradient(al, tl, ar, tr):
    return lambda X: jnp.where(X <= tl, al,
                     jnp.where(X < tr, 1, ar))


class ActivationFunction(object):
    def __call__(self, X: Matrix) -> Matrix:
        raise NotImplementedError

    def gradient(self, X: Matrix) -> Matrix:
        raise NotImplementedError


class ReLUActivation(ActivationFunction):
    def __call__(self, X: Matrix) -> Matrix:
        return Relu(X)

    def gradient(self, X: Matrix) -> Matrix:
        return Relu_gradient(X)


class LeakyReLUActivation(ActivationFunction):
    def __init__(self, alpha):
        self.alpha = alpha

    def __call__(self, X: Matrix) -> Matrix:
        return Leaky_relu(self.alpha)(X)

    def gradient(self, X: Matrix) -> Matrix:
        return Leaky_relu_gradient(self.alpha)(X)


class AllReLUActivation(ActivationFunction):
    def __init__(self, alpha):
        self.alpha = alpha

    def __call__(self, X: Matrix) -> Matrix:
        return All_relu(self.alpha)(X)

    def gradient(self, X: Matrix) -> Matrix:
        return All_relu_gradient(self.alpha)(X)


class HyperbolicTangentActivation(ActivationFunction):
    def __call__(self, X: Matrix) -> Matrix:
        return Hyperbolic_tangent(X)

    def gradient(self, X: Matrix) -> Matrix:
        return Hyperbolic_tangent_gradient(X)


class SigmoidActivation(ActivationFunction):
    def __call__(self, X: Matrix) -> Matrix:
        return Sigmoid(X)

    def gradient(self, X: Matrix) -> Matrix:
        return Sigmoid_gradient(X)


class SReLUActivation(ActivationFunction):
    def __init__(self, al=0.0, tl=0.0, ar=0.0, tr=1.0):
        # Store the parameters and their gradients in matrices.
        # This is to make them usable for optimizers.
        self.x = jnp.array([al, tl, ar, tr])
        self.Dx = jnp.array([0.0, 0.0, 0.0, 0.0])

    def __call__(self, X: Matrix) -> Matrix:
        al, tl, ar, tr = self.x
        return Srelu(al, tl, ar, tr)(X)

    def gradient(self, X: Matrix) -> Matrix:
        al, tl, ar, tr = self.x
        return Srelu_gradient(al, tl, ar, tr)(X)


def parse_activation(text: str) -> ActivationFunction:
    try:
        func = parse_function_call(text)
        if func.name == 'ReLU':
            return ReLUActivation()
        elif func.name == 'Sigmoid':
            return SigmoidActivation()
        elif func.name == 'HyperbolicTangent':
            return HyperbolicTangentActivation()
        elif func.name == 'AllReLU':
            alpha = func.as_scalar('alpha')
            return AllReLUActivation(alpha)
        elif func.name == 'LeakyReLU':
            alpha = func.as_scalar('alpha')
            return LeakyReLUActivation(alpha)
        elif func.name == 'SReLU':
            al = func.as_scalar('al', 0)
            tl = func.as_scalar('tl', 0)
            ar = func.as_scalar('ar', 0)
            tr = func.as_scalar('tr', 1)
            return SReLUActivation(al, tl, ar, tr)
    except:
        pass
    raise RuntimeError(f'Could not parse activation "{text}"')



# Copyright 2023 Wieger Wesselink.
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from pathlib import Path
from typing import Tuple

import jax.numpy as jnp
import numpy as np


def to_one_hot(x: jnp.ndarray, n_classes: int):
    one_hot = np.zeros((len(x), n_classes), dtype=float)
    one_hot[np.arange(len(x)), x] = 1
    return jnp.array(one_hot)


class MemoryDataLoader(object):
    """
    A data loader with an interface similar to torch.utils.data.DataLoader.
    """

    def __init__(self, Xdata: jnp.ndarray, Tdata: jnp.ndarray, batch_size: int=True, num_classes=0):
        """
        :param Xdata: a dataset with row layout
        :param Tdata: the expected targets. In case of a classification task the targets may be specified as a vector
                      of integers that denote the expected classes. In such a case the targets will be expanded on the
                      fly using one hot encoding.
        :param batch_size: the batch size
        :param num_classes: the number of classes in case of a classification problem, 0 otherwise
        """
        self.Xdata = Xdata
        self.Tdata = Tdata
        self.batch_size = batch_size
        self.dataset = Xdata
        self.num_classes = int(Tdata.max() + 1) if num_classes == 0 and len(Tdata.shape) == 1 else num_classes

    def __iter__(self):
        N = self.Xdata.shape[0]  # N is the number of examples
        K = N // self.batch_size  # K is the number of batches
        for k in range(K):
            batch = jnp.array(range(k * self.batch_size, (k + 1) * self.batch_size))
            Xbatch = self.Xdata[batch]
            Tbatch = self.Tdata[batch]
            yield Xbatch, to_one_hot(Tbatch, self.num_classes) if self.num_classes else Tbatch
    def __len__(self):
        """
        Returns the number of batches
        """
        return self.Xdata.shape[0] // self.batch_size


DataLoader = MemoryDataLoader


def create_npz_dataloaders(filename: str, batch_size: int=True) -> Tuple[MemoryDataLoader, MemoryDataLoader]:
    """
    Creates a data loader from a file containing a dictionary with Xtrain, Ttrain, Xtest and Ttest tensors
    :param filename: a file in NumPy .npz format
    :param batch_size: the batch size of the data loader
    :return: a tuple of data loaders
    """
    path = Path(filename)
    print(f'Loading dataset from file {path}')
    if not path.exists():
        raise RuntimeError(f"Could not load file '{path}'")

    data = dict(jnp.load(filename, allow_pickle=True))
    Xtrain, Ttrain, Xtest, Ttest = data['Xtrain'], data['Ttrain'], data['Xtest'], data['Ttest']
    train_loader = MemoryDataLoader(jnp.array(Xtrain), jnp.array(Ttrain), batch_size)
    test_loader = MemoryDataLoader(jnp.array(Xtest), jnp.array(Ttest), batch_size)
    return train_loader, test_loader


# Copyright 2023 Wieger Wesselink.
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE or http://www.boost.org/LICENSE_1_0.txt)

import jax.numpy as jnp
from nerva_jax.matrix_operations import Diag, column_repeat, exp, hadamard, identity, is_row_vector, log, reciprocal, \
    row_repeat, rows_max, rows_sum

Matrix = jnp.ndarray

def softmax(X: Matrix) -> Matrix:
    N, D = X.shape
    E = exp(X)
    return hadamard(E, column_repeat(reciprocal(rows_sum(E)), D))


def softmax_jacobian(x: Matrix) -> Matrix:
    assert is_row_vector(x)
    y = softmax(x)
    return Diag(y) - y.T * y


def stable_softmax(X: Matrix) -> Matrix:
    N, D = X.shape
    Y = X - column_repeat(rows_max(X), D)
    E = exp(Y)
    return hadamard(E, column_repeat(reciprocal(rows_sum(E)), D))


def stable_softmax_jacobian(x: Matrix) -> Matrix:
    assert is_row_vector(x)
    y = stable_softmax(x)
    return Diag(y) - y.T * y


def log_softmax(X: Matrix) -> Matrix:
    N, D = X.shape
    return X - column_repeat(log(rows_sum(exp(X))), D)


def log_softmax_jacobian(x: Matrix) -> Matrix:
    assert is_row_vector(x)
    N, D = x.shape
    return identity(D) - row_repeat(softmax(x), D)


def stable_log_softmax(X: Matrix) -> Matrix:
    N, D = X.shape
    Y = X - column_repeat(rows_max(X), D)
    return Y - column_repeat(log(rows_sum(exp(Y))), D)


def stable_log_softmax_jacobian(x: Matrix) -> Matrix:
    return log_softmax_jacobian(x)

Matrix = jnp.ndarray

class SoftmaxFunction(object):
    def __call__(self, X: Matrix) -> Matrix:
        return softmax(X)

    def jacobian(self, X: Matrix) -> Matrix:
        return softmax_jacobian(X)


class StableSoftmaxFunction(object):
    def __call__(self, X: Matrix) -> Matrix:
        return stable_softmax(X)

    def jacobian(self, X: Matrix) -> Matrix:
        return stable_softmax_jacobian(X)


class LogSoftmaxFunction(object):
    def __call__(self, X: Matrix) -> Matrix:
        return log_softmax(X)

    def jacobian(self, X: Matrix) -> Matrix:
        return log_softmax_jacobian(X)


class StableLogSoftmaxFunction(object):
    def __call__(self, X: Matrix) -> Matrix:
        return stable_log_softmax(X)

    def jacobian(self, X: Matrix) -> Matrix:
        return stable_log_softmax_jacobian(X)


# Copyright 2023 Wieger Wesselink.
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE or http://www.boost.org/LICENSE_1_0.txt)

import re
import time
from typing import Dict

import jax.numpy as jnp


def set_jax_options():
    jnp.set_printoptions(precision=8, edgeitems=3, threshold=5, suppress=True, linewidth=160)


def pp(name: str, x: jnp.ndarray):
    if x.ndim == 1:
        print(f'{name} ({x.shape[0]})\n{x}')
    else:
        print(f'{name} ({x.shape[0]}x{x.shape[1]})\n{x}')


class StopWatch(object):
    def __init__(self):
        self.start = time.perf_counter()

    def seconds(self):
        end = time.perf_counter()
        return end - self.start

    def reset(self):
        self.start = time.perf_counter()


class FunctionCall:
    def __init__(self, name: str, arguments: dict):
        self.name = name
        self.arguments = arguments

    def has_key(self, key: str) -> bool:
        return key in self.arguments

    def get_value(self, key: str) -> str:
        if key in self.arguments:
            return self.arguments[key]
        elif len(self.arguments) == 1 and "" in self.arguments:
            return self.arguments[""]
        return ""

    def as_scalar(self, key: str, default_value: float = None) -> float:
        value = self.get_value(key)
        if value:
            return float(value)  # Assuming scalar is a float, adjust if necessary
        elif default_value:
            return default_value
        raise RuntimeError(f"Could not find an argument named \"{key}\"")

    def as_string(self, key: str, default_value: str = "") -> str:
        value = self.get_value(key)
        if value:
            return value
        elif default_value:
            return default_value
        raise RuntimeError(f"Could not find an argument named \"{key}\"")


def parse_function_call(text: str) -> FunctionCall:
    """
    Parse a string of the shape `NAME(key1=value1, key2=value2, ...)`.
    If there are no arguments the parentheses may be omitted.
    If there is only one parameter, it is allowed to pass `NAME(value)` instead of `NAME(key=value)`
    """

    text = text.strip()

    def error():
        raise RuntimeError(f"Could not parse function call \"{text}\"")

    name = ""
    arguments = {}

    # no parentheses
    match = re.match(r"(\w+$)", text)
    if match:
        name = match.group(1)
        return FunctionCall(name, arguments)

    # with parentheses
    match = re.match(r"(\w+)\((.*?)\)", text)
    if match:
        name = match.group(1)
        args = re.split(r",", match.group(2))

        if len(args) == 1 and "=" not in args[0]:
            # NAME(value)
            value = args[0].strip()
            arguments[""] = value
            return FunctionCall(name, arguments)
        else:
            # NAME(key1=value1, ...)
            for arg in args:
                words = re.split(r"\s*=\s*", arg.strip())
                if len(words) != 2:
                    error()
                key, value = words
                if key in arguments:
                    print(f"Key \"{key}\" appears multiple times.")
                    error()
                arguments[key] = value
            return FunctionCall(name, arguments)

    error()


def load_dict_from_npz(filename: str) -> Dict[str, jnp.ndarray]:
    """
    Loads a dictionary from a file in .npz format
    :param filename: a file name
    :return: a dictionary
    """

    return dict(jnp.load(filename, allow_pickle=True))



# Copyright 2023 Wieger Wesselink.
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from typing import List

import jax.numpy as jnp
from nerva_jax.layers import BatchNormalizationLayer, LinearLayer, parse_linear_layer
from nerva_jax.utilities import load_dict_from_npz, pp

Matrix = jnp.ndarray


class MultilayerPerceptron(object):
    """
    Multilayer perceptron
    """
    def __init__(self, layers=None):
        if not layers:
            layers = []
        self.layers = layers

    def feedforward(self, X: Matrix) -> Matrix:
        for layer in self.layers:
            X = layer.feedforward(X)
        return X

    def backpropagate(self, Y: Matrix, DY: Matrix) -> None:
        for layer in reversed(self.layers):
            layer.backpropagate(Y, DY)
            Y, DY = layer.X, layer.DX

    def optimize(self, eta: float):
        for layer in self.layers:
            layer.optimize(eta)

    def info(self):
        index = 1
        for layer in self.layers:
            if isinstance(layer, LinearLayer):
                pp(f'W{index}', layer.W)
                pp(f'b{index}', layer.b)
                index += 1

    def load_weights_and_bias(self, filename: str):
        """
        Loads the weights and biases from a file in .npz format

        The weight matrices are stored using the keys W1, W2, ... and the bias vectors using the keys b1, b2, ...
        :param filename: the name of the file
        """
        print(f'Loading weights and bias from {filename}')
        data = load_dict_from_npz(filename)
        index = 1
        for layer in self.layers:
            if isinstance(layer, LinearLayer):
                layer.W = jnp.array(data[f'W{index}'])
                layer.b = jnp.array(data[f'b{index}'])
                index += 1


def parse_multilayer_perceptron(layer_specifications: List[str],
                                linear_layer_sizes: List[int],
                                optimizers: List[str],
                                linear_layer_weight_initializers: List[str]
                               ) -> MultilayerPerceptron:

    assert len(linear_layer_weight_initializers) == len(linear_layer_sizes) - 1
    layers = []

    linear_layer_index = 0
    optimizer_index = 0
    D = linear_layer_sizes[linear_layer_index]  # the input size of the current layer

    for specification in layer_specifications:
        if specification == 'BatchNormalization':
            layer = BatchNormalizationLayer(D)
            optimizer = optimizers[optimizer_index]
            layer.set_optimizer(optimizer)
            optimizer_index += 1
        else:
            K = linear_layer_sizes[linear_layer_index + 1]  # the output size of the layer
            optimizer = optimizers[optimizer_index]
            weight_initializer = linear_layer_weight_initializers[linear_layer_index]
            layer = parse_linear_layer(specification, D, K, optimizer, weight_initializer)
            optimizer_index += 1
            linear_layer_index += 1
            D = K
        layers.append(layer)
    return MultilayerPerceptron(layers)



# Copyright 2023 Wieger Wesselink.
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from typing import Any, Callable, List

import jax.numpy as jnp
from nerva_jax.utilities import parse_function_call


class Optimizer(object):
    def update(self, eta):
        raise NotImplementedError


class CompositeOptimizer(Optimizer):
    def __init__(self, optimizers: List[Optimizer]):
        self.optimizers = optimizers

    def update(self, eta):
        for optimizer in self.optimizers:
            optimizer.update(eta)


class GradientDescentOptimizer(Optimizer):
    def __init__(self, obj, attr_x: str, attr_Dx: str):
        """
        Store the names of the x and Dx attributes
        """
        self.obj = obj
        self.attr_x = attr_x
        self.attr_Dx = attr_Dx

    def update(self, eta):
        x = getattr(self.obj, self.attr_x)
        Dx = getattr(self.obj, self.attr_Dx)
        x1 = x - eta * Dx
        setattr(self.obj, self.attr_x, x1)


class MomentumOptimizer(GradientDescentOptimizer):
    def __init__(self, obj, attr_x: str, attr_Dx: str, mu: float):
        super().__init__(obj, attr_x, attr_Dx)
        self.mu = mu
        x = getattr(self.obj, self.attr_x)
        self.delta_x = jnp.zeros_like(x)

    def update(self, eta):
        x = getattr(self.obj, self.attr_x)
        Dx = getattr(self.obj, self.attr_Dx)
        self.delta_x = self.mu * self.delta_x - eta * Dx
        x1 = x + self.delta_x
        setattr(self.obj, self.attr_x, x1)

class NesterovOptimizer(MomentumOptimizer):
    def __init__(self, obj, attr_x: str, attr_Dx: str, mu: float):
        super().__init__(obj, attr_x, attr_Dx, mu)

    def update(self, eta):
        x = getattr(self.obj, self.attr_x)
        Dx = getattr(self.obj, self.attr_Dx)
        self.delta_x_prev = self.delta_x
        self.delta_x = self.mu * self.delta_x - eta * Dx
        x1 = x + self.mu * self.delta_x - eta * Dx
        setattr(self.obj, self.attr_x, x1)


def parse_optimizer(text: str) -> Callable[[Any, Any], Optimizer]:
    try:
        func = parse_function_call(text)
        if func.name == 'GradientDescent':
            return lambda obj, attr_x, attr_Dx: GradientDescentOptimizer(obj, attr_x, attr_Dx)
        elif func.name == 'Momentum':
            mu = func.as_scalar('mu')
            return lambda obj, attr_x, attr_Dx: MomentumOptimizer(obj, attr_x, attr_Dx, mu)
        elif func.name == 'Nesterov':
            mu = func.as_scalar('mu')
            return lambda obj, attr_x, attr_Dx: NesterovOptimizer(obj, attr_x, attr_Dx, mu)
    except:
        pass
    raise RuntimeError(f'Could not parse optimizer "{text}"')



# Copyright 2023 Wieger Wesselink.
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE or http://www.boost.org/LICENSE_1_0.txt)

import jax
import jax.numpy as jnp

Matrix = jax.numpy.ndarray


# A constant used by inv_sqrt to avoid division by zero
epsilon = 1e-7


def is_vector(x: Matrix) -> bool:
    return len(x.shape) == 1


def is_column_vector(x: Matrix) -> bool:
    return is_vector(x) or x.shape[1] == 1


def is_row_vector(x: Matrix) -> bool:
    return is_vector(x) or x.shape[0] == 1


def vector_size(x: Matrix) -> int:
    return x.shape[0]


def is_square(X: Matrix) -> bool:
    m, n = X.shape
    return m == n


def dot(x: Matrix, y: Matrix):
    return jnp.dot(jnp.squeeze(x), jnp.squeeze(y))


def zeros(m: int, n=None) -> Matrix:
    """
    Returns an mxn matrix with all elements equal to 0.
    """
    return jnp.zeros((m, n)) if n else jnp.zeros(m)


def ones(m: int, n=None) -> Matrix:
    """
    Returns an mxn matrix with all elements equal to 1.
    """
    return jnp.ones((m, n)) if n else jnp.ones(m)


def identity(n: int) -> Matrix:
    """
    Returns the nxn identity matrix.
    """
    return jnp.eye(n)


def product(X: Matrix, Y: Matrix) -> Matrix:
    return X @ Y


def hadamard(X: Matrix, Y: Matrix) -> Matrix:
    return X * Y


def diag(X: Matrix) -> Matrix:
    return jnp.diag(X)


def Diag(x: Matrix) -> Matrix:
    return jnp.diagflat(x)


def elements_sum(X: Matrix):
    """
    Returns the sum of the elements of X.
    """
    return jnp.sum(X)


def column_repeat(x: Matrix, n: int) -> Matrix:
    assert is_column_vector(x)
    if len(x.shape) == 1:
        x = x[:, jnp.newaxis]
    return jnp.tile(x, (1, n))


def row_repeat(x: Matrix, m: int) -> Matrix:
    assert is_row_vector(x)
    if len(x.shape) == 1:
        x = x[jnp.newaxis, :]
    return jnp.tile(x, (m, 1))


def columns_sum(X: Matrix) -> Matrix:
    return jnp.sum(X, axis=0)


def rows_sum(X: Matrix) -> Matrix:
    return jnp.sum(X, axis=1)


def columns_max(X: Matrix) -> Matrix:
    """
    Returns a column vector with the maximum values of each row in X.
    """
    return jnp.max(X, axis=0)


def rows_max(X: Matrix) -> Matrix:
    """
    Returns a row vector with the maximum values of each column in X.
    """
    return jnp.max(X, axis=1)


def columns_mean(X: Matrix) -> Matrix:
    """
    Returns a column vector with the mean values of each row in X.
    """
    return jnp.mean(X, axis=0)


def rows_mean(X: Matrix) -> Matrix:
    """
    Returns a row vector with the mean values of each column in X.
    """
    return jnp.mean(X, axis=1)


def apply(f, X: Matrix) -> Matrix:
    return f(X)


def exp(X: Matrix) -> Matrix:
    return jnp.exp(X)


def log(X: Matrix) -> Matrix:
    return jnp.log(X)


def reciprocal(X: Matrix) -> Matrix:
    return 1 / X


def square(X: Matrix) -> Matrix:
    return jnp.square(X)


def sqrt(X: Matrix) -> Matrix:
    return jnp.sqrt(X)


def inv_sqrt(X: Matrix) -> Matrix:
    return 1 / jnp.sqrt(X + epsilon)  # The epsilon is needed for numerical stability


def log_sigmoid(X: Matrix) -> Matrix:
    return -jnp.logaddexp(0, -X)



# Copyright 2023 Wieger Wesselink.
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE or http://www.boost.org/LICENSE_1_0.txt)

import jax.numpy as jnp
from nerva_jax.activation_functions import Sigmoid
from nerva_jax.matrix_operations import column_repeat, dot, elements_sum, hadamard, log, log_sigmoid, reciprocal, \
    rows_sum
from nerva_jax.softmax_functions import log_softmax, softmax, stable_log_softmax, stable_softmax


# Naming conventions:
# - lowercase functions operate on vectors (y and t)
# - uppercase functions operate on matrices (Y and T)


def squared_error_loss(y, t):
    return dot(y - t, y - t)


def squared_error_loss_gradient(y, t):
    return 2 * (y - t)


def Squared_error_loss(Y, T):
    return elements_sum(hadamard(Y - T, Y - T))


def Squared_error_loss_gradient(Y, T):
    return 2 * (Y - T)


def mean_squared_error_loss(y, t):
    N, K = y.shape
    return squared_error_loss(y, t) / K


def mean_squared_error_loss_gradient(y, t):
    N, K = y.shape
    return squared_error_loss_gradient(y, t) / K


def Mean_squared_error_loss(Y, T):
    N, K = Y.shape
    return Squared_error_loss(Y, T) / (K * N)


def Mean_squared_error_loss_gradient(Y, T):
    N, K = Y.shape
    return Squared_error_loss_gradient(Y, T) / (K * N)


def cross_entropy_loss(y, t):
    return -dot(t, log(y))


def cross_entropy_loss_gradient(y, t):
    return -hadamard(t, reciprocal(y))


def Cross_entropy_loss(Y, T):
    return -elements_sum(hadamard(T, log(Y)))


def Cross_entropy_loss_gradient(Y, T):
    return -hadamard(T, reciprocal(Y))


def softmax_cross_entropy_loss(y, t):
    return -dot(t, log_softmax(y))


def softmax_cross_entropy_loss_gradient(y, t):
    return elements_sum(t) * softmax(y) - t


def softmax_cross_entropy_loss_gradient_one_hot(y, t):
    return softmax(y) - t


def Softmax_cross_entropy_loss(Y, T):
    return -elements_sum(hadamard(T, log_softmax(Y)))


def Softmax_cross_entropy_loss_gradient(Y, T):
    N, K = Y.shape
    return hadamard(softmax(Y), column_repeat(rows_sum(T), K)) - T

def Softmax_cross_entropy_loss_gradient_one_hot(Y, T):
    return softmax(Y) - T


def stable_softmax_cross_entropy_loss(y, t):
    return -dot(t, stable_log_softmax(y))


def stable_softmax_cross_entropy_loss_gradient(y, t):
    return stable_softmax(y) * elements_sum(t) - t


def stable_softmax_cross_entropy_loss_gradient_one_hot(y, t):
    return stable_softmax(y) - t


def Stable_softmax_cross_entropy_loss(Y, T):
    return -elements_sum(hadamard(T, stable_log_softmax(Y)))


def Stable_softmax_cross_entropy_loss_gradient(Y, T):
    N, K = Y.shape
    return hadamard(stable_softmax(Y), column_repeat(rows_sum(T), K)) - T


def Stable_softmax_cross_entropy_loss_gradient_one_hot(Y, T):
    return stable_softmax(Y) - T


def logistic_cross_entropy_loss(y, t):
    return -dot(t, log_sigmoid(y))


def logistic_cross_entropy_loss_gradient(y, t):
    return hadamard(t, Sigmoid(y)) - t


def Logistic_cross_entropy_loss(Y, T):
    return -elements_sum(hadamard(T, log_sigmoid(Y)))


def Logistic_cross_entropy_loss_gradient(Y, T):
    return hadamard(T, Sigmoid(Y)) - T


def negative_log_likelihood_loss(y, t):
    return -log(dot(y, t))


def negative_log_likelihood_loss_gradient(y, t):
    return -reciprocal(dot(y, t)) * t


def Negative_log_likelihood_loss(Y, T):
    return -elements_sum(log(rows_sum(hadamard(Y, T))))


def Negative_log_likelihood_loss_gradient(Y, T):
    N, K = Y.shape
    return -hadamard(column_repeat(reciprocal(rows_sum(hadamard(Y, T))), K), T)

Matrix = jnp.ndarray


class LossFunction(object):
    def __call__(self, Y: Matrix, T: Matrix) -> float:
        raise NotImplementedError

    def gradient(self, Y: Matrix, T: Matrix) -> Matrix:
        raise NotImplementedError


class SquaredErrorLossFunction(LossFunction):
    def __call__(self, Y: Matrix, T: Matrix) -> float:
        return Squared_error_loss(Y, T)

    def gradient(self, Y: Matrix, T: Matrix) -> Matrix:
        return Squared_error_loss_gradient(Y, T)


class MeanSquaredErrorLossFunction(LossFunction):
    def __call__(self, Y: Matrix, T: Matrix) -> float:
        return Mean_squared_error_loss(Y, T)

    def gradient(self, Y: Matrix, T: Matrix) -> Matrix:
        return Mean_squared_error_loss_gradient(Y, T)


class CrossEntropyLossFunction(LossFunction):
    def __call__(self, Y: Matrix, T: Matrix) -> float:
        return Cross_entropy_loss(Y, T)

    def gradient(self, Y: Matrix, T: Matrix) -> Matrix:
        return Cross_entropy_loss_gradient(Y, T)


class SoftmaxCrossEntropyLossFunction(LossFunction):
    def __call__(self, Y: Matrix, T: Matrix) -> float:
        return Softmax_cross_entropy_loss(Y, T)

    def gradient(self, Y: Matrix, T: Matrix) -> Matrix:
        return Softmax_cross_entropy_loss_gradient(Y, T)


class StableSoftmaxCrossEntropyLossFunction(LossFunction):
    def __call__(self, Y: Matrix, T: Matrix) -> float:
        return Stable_softmax_cross_entropy_loss(Y, T)

    def gradient(self, Y: Matrix, T: Matrix) -> Matrix:
        return Stable_softmax_cross_entropy_loss_gradient(Y, T)


class LogisticCrossEntropyLossFunction(LossFunction):
    def __call__(self, Y: Matrix, T: Matrix) -> float:
        return Logistic_cross_entropy_loss(Y, T)

    def gradient(self, Y: Matrix, T: Matrix) -> Matrix:
        return Logistic_cross_entropy_loss_gradient(Y, T)


class NegativeLogLikelihoodLossFunction(LossFunction):
    def __call__(self, Y: Matrix, T: Matrix) -> float:
        return Negative_log_likelihood_loss(Y, T)

    def gradient(self, Y: Matrix, T: Matrix) -> Matrix:
        return Negative_log_likelihood_loss_gradient(Y, T)


def parse_loss_function(text: str) -> LossFunction:
    if text == "SquaredError":
        return SquaredErrorLossFunction()
    elif text == "MeanSquaredError":
        return MeanSquaredErrorLossFunction()
    elif text == "CrossEntropy":
        return CrossEntropyLossFunction()
    elif text == "SoftmaxCrossEntropy":
        return StableSoftmaxCrossEntropyLossFunction()
    elif text == "LogisticCrossEntropy":
        return LogisticCrossEntropyLossFunction()
    elif text == "NegativeLogLikelihood":
        return NegativeLogLikelihoodLossFunction()
    else:
        raise RuntimeError(f"unknown loss function '{text}'")


# Copyright 2023 Wieger Wesselink.
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE or http://www.boost.org/LICENSE_1_0.txt)


from typing import List

from nerva_jax.datasets import DataLoader, create_npz_dataloaders
from nerva_jax.learning_rate import LearningRateScheduler, parse_learning_rate
from nerva_jax.loss_functions import LossFunction, parse_loss_function
from nerva_jax.multilayer_perceptron import MultilayerPerceptron, parse_multilayer_perceptron
from nerva_jax.utilities import StopWatch, pp, set_jax_options


class SGDOptions(object):
    debug=False


def print_epoch(epoch, lr, loss, train_accuracy, test_accuracy, elapsed):
    print(f'epoch {epoch:3}  '
          f'lr: {lr:.8f}  '
          f'loss: {loss:.8f}  '
          f'train accuracy: {train_accuracy:.8f}  '
          f'test accuracy: {test_accuracy:.8f}  '
          f'time: {elapsed:.8f}s'
         )

def compute_accuracy(M: MultilayerPerceptron, data_loader: DataLoader):
    N = len(data_loader.dataset)  # N is the number of examples
    total_correct = 0
    for X, T in data_loader:
        Y = M.feedforward(X)
        predicted = Y.argmax(axis=1)  # the predicted classes for the batch
        targets = T.argmax(axis=1)    # the expected classes
        total_correct += (predicted == targets).sum().item()

    return total_correct / N


def compute_loss(M: MultilayerPerceptron, data_loader: DataLoader, loss: LossFunction):
    N = len(data_loader.dataset)  # N is the number of examples
    total_loss = 0.0
    for X, T in data_loader:
        Y = M.feedforward(X)
        total_loss += loss(Y, T)

    return total_loss / N


def compute_statistics(M, lr, loss, train_loader, test_loader, epoch, elapsed_seconds=0.0, print_statistics=True):
    if print_statistics:
        train_loss = compute_loss(M, train_loader, loss)
        train_accuracy = compute_accuracy(M, train_loader)
        test_accuracy = compute_accuracy(M, test_loader)
        print_epoch(epoch, lr, train_loss, train_accuracy, test_accuracy, elapsed_seconds)
    else:
        print(f'epoch {epoch:3}')


def sgd(M: MultilayerPerceptron,
        epochs: int,
        loss: LossFunction,
        learning_rate: LearningRateScheduler,
        train_loader: DataLoader,
        test_loader: DataLoader
       ):

    lr = learning_rate(0)
    compute_statistics(M, lr, loss, train_loader, test_loader, epoch=0)
    training_time = 0.0

    for epoch in range(epochs):
        timer = StopWatch()
        lr = learning_rate(epoch)  # update the learning at the start of each epoch

        for k, (X, T) in enumerate(train_loader):
            Y = M.feedforward(X)
            DY = loss.gradient(Y, T) / Y.shape[0]

            if SGDOptions.debug:
                print(f'epoch: {epoch} batch: {k}')
                M.info()
                pp("X", X)
                pp("Y", Y)
                pp("DY", DY)

            M.backpropagate(Y, DY)
            M.optimize(lr)

        seconds = timer.seconds()
        training_time += seconds
        compute_statistics(M, lr, loss, train_loader, test_loader, epoch=epoch + 1, elapsed_seconds=seconds)

    print(f'Total training time for the {epochs} epochs: {training_time:.8f}s\n')


def train(layer_specifications: List[str],
          linear_layer_sizes: List[int],
          linear_layer_optimizers: List[str],
          linear_layer_weight_initializers: List[str],
          batch_size: int,
          epochs: int,
          loss: str,
          learning_rate: str,
          weights_and_bias_file: str,
          dataset_file: str,
          debug: bool
         ):
    SGDOptions.debug = debug
    set_jax_options()
    loss = parse_loss_function(loss)
    learning_rate = parse_learning_rate(learning_rate)
    M = parse_multilayer_perceptron(layer_specifications, linear_layer_sizes, linear_layer_optimizers, linear_layer_weight_initializers)
    if weights_and_bias_file:
        M.load_weights_and_bias(weights_and_bias_file)
    train_loader, test_loader = create_npz_dataloaders(dataset_file, batch_size=batch_size)
    sgd(M, epochs, loss, learning_rate, train_loader, test_loader)
